# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT

import shutil
import subprocess

import yaml
import requests

from envs_manager.api import EnvManagerInstance

MICROMAMBA_VARIANT = "micromamba"
CONDA_VARIANT = "conda"
ANACONDA_API_PACKAGE_INFO = "https://api.anaconda.org/package/{channel}/{package_name}"


class CondaLikeInterface(EnvManagerInstance):
    ID = "conda-like"

    def _get_package_info(self, package_name, channel):
        if not channel:
            channel = "anaconda"
        package_info_url = ANACONDA_API_PACKAGE_INFO.format(
            channel=channel, package_name=package_name
        )
        package_info = requests.get(package_info_url).json()
        return package_info

    def validate(self):
        if self.external_executable:
            command = [self.external_executable, "--version"]
            try:
                result = subprocess.run(
                    command, capture_output=True, check=True, text=True
                )
                version = result.stdout.split()
                if len(version) <= 1:
                    self.executable_variant = MICROMAMBA_VARIANT
                else:
                    self.executable_variant = version[0]
                return True
            except Exception as error:
                print(error.stderr)
        return False

    def create_environment(self, packages=[], channels=["conda-forge"]):
        command = [self.external_executable, "create", "-p", self.environment_path]
        if packages:
            command += packages
        if channels:
            command += ["-c"] + channels
        try:
            result = subprocess.run(
                command, stderr=subprocess.PIPE, check=True, text=True
            )
            print(result.stdout)
            return (True, result)
        except Exception as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def delete_environment(self):
        command = [
            self.external_executable,
            "remove",
            "-p",
            self.environment_path,
            "--all",
            "-y",
        ]
        try:
            result = subprocess.run(command, capture_output=True, check=True, text=True)
            print(result.stdout)
            shutil.rmtree(self.environment_path, ignore_errors=True)
            return (True, result)
        except Exception as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def activate_environment(self):
        raise NotImplementedError()

    def deactivate_environment(self):
        raise NotImplementedError()

    def export_environment(self, export_file_path=None):
        command = [
            self.external_executable,
            "env",
            "export",
            "-p",
            self.environment_path,
            "--from-history",
        ]
        try:
            result = subprocess.run(command, capture_output=True, check=True, text=True)
            if export_file_path:
                with open(export_file_path, "w") as exported_file:
                    exported_file.write(result.stdout)
                print(result.stdout)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def import_environment(self, import_file_path):
        if self.executable_variant == MICROMAMBA_VARIANT:
            command = [
                self.external_executable,
                "create",
                "-p",
                self.environment_path,
                f"--file={import_file_path}",
            ]
        else:
            command = [
                self.external_executable,
                "env",
                "create",
                "-p",
                self.environment_path,
                f"--file={import_file_path}",
            ]
        try:
            result = subprocess.run(command, capture_output=True, check=True, text=True)
            print(result.stdout)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (
                False,
                f"{error.returncode}: {error.stderr}\nNote: Importing environments only works for environment definitions created with the same operating system.",
            )

    def install_packages(
        self,
        packages,
        channels=["conda-forge"],
        force=False,
        capture_output=False,
    ):
        if self.executable_variant != MICROMAMBA_VARIANT:
            packages = [f"'{package}'" for package in packages]
        command = [
            self.external_executable,
            "install",
            "-p",
            self.environment_path,
        ] + packages
        if force:
            command += ["-y"]
        if channels:
            channels = ["-c"] + channels
            command += channels
        try:
            if capture_output:
                result = subprocess.run(
                    command, capture_output=True, check=True, text=True
                )
            else:
                result = subprocess.run(
                    command, stderr=subprocess.PIPE, check=True, text=True
                )
            return (True, result)
        except subprocess.CalledProcessError as error:
            formatted_error = f"{error.returncode}: {error.stderr}"
            return (False, formatted_error)

    def uninstall_packages(self, packages, force=False, capture_output=False):
        command = [
            self.external_executable,
            "remove",
            "-p",
            self.environment_path,
        ] + packages
        if force:
            command + ["-y"]
        try:
            if capture_output:
                result = subprocess.run(
                    command, capture_output=capture_output, check=True, text=True
                )
            else:
                result = subprocess.run(
                    command, stderr=subprocess.PIPE, check=True, text=True
                )
            return (True, result)
        except subprocess.CalledProcessError as error:
            if "PackagesNotFoundError" in error.stderr:
                return (True, error)
            formatted_error = f"{error.returncode}: {error.stderr}"
            return (False, formatted_error)

    def update_packages(self, packages, force=False, capture_output=False):
        command = [
            self.external_executable,
            "update",
            "-p",
            self.environment_path,
        ] + packages
        if force:
            command + ["-y"]
        try:
            if capture_output:
                result = subprocess.run(
                    command, capture_output=capture_output, check=True, text=True
                )
                if "All requested packages already installed" in result.stdout:
                    return (False, result.stdout)
            else:
                result = subprocess.run(
                    command, stderr=subprocess.PIPE, check=True, text=True
                )
            return (True, result)
        except subprocess.CalledProcessError as error:
            formatted_error = f"{error.returncode}: {error.stderr}"
            return (False, formatted_error)

    def list_packages(self):
        command = [self.external_executable, "list", "-p", self.environment_path]
        result = subprocess.run(command, capture_output=True, check=True, text=True)
        result_lines = result.stdout.split("\n")
        export_env_result, export_env_data = self.export_environment()
        if export_env_result:
            packages_requested = yaml.load(
                export_env_data.stdout, Loader=yaml.FullLoader
            )["dependencies"]
        else:
            packages_requested = []

        if self.executable_variant == MICROMAMBA_VARIANT:
            skip_lines = 4
        else:
            skip_lines = 3
        formatted_packages = {}
        formatted_list = dict(
            environment=self.environment_path, packages=formatted_packages
        )

        for package in result_lines[skip_lines:-1]:
            package_info = package.split()
            package_name = package_info[0]
            package_build = None if len(package_info) <= 2 else package_info[2]
            package_channel = None if len(package_info) <= 3 else package_info[3]
            if package_channel:
                package_description = self._get_package_info(
                    package_name, channel=package_channel
                )["summary"]
            else:
                package_description = None
            package_requested = package_name in packages_requested
            formatted_package = dict(
                name=package_name,
                version=package_info[1],
                build=package_build,
                channel=package_channel,
                description=package_description,
                requested=package_requested,
            )
            formatted_packages[package_name] = formatted_package

        print(result.stdout)
        return formatted_list



# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT

import os
from pathlib import Path

from envs_manager.backends.venv_interface import VEnvInterface
from envs_manager.backends.conda_like_interface import CondaLikeInterface


DEFAULT_BACKENDS_ROOT_PATH = Path(
    os.environ.get(
        "BACKENDS_ROOT_PATH", str(Path.home() / ".envs-manager" / "backends")
    )
)
DEFAULT_BACKEND = os.environ.get("ENV_BACKEND", "venv")
DEFAULT_ENVS_ROOT_PATH = DEFAULT_BACKENDS_ROOT_PATH / DEFAULT_BACKEND / "envs"
EXTERNAL_EXECUTABLE = os.environ.get("ENV_BACKEND_EXECUTABLE", None)


class Manager:
    """
    Class to handle different Python environment and packages managers implementations.
    """

    BACKENDS = {
        VEnvInterface.ID: VEnvInterface,
        CondaLikeInterface.ID: CondaLikeInterface,
    }

    def __init__(
        self,
        backend,
        root_path=None,
        env_name=None,
        env_directory=None,
        external_executable=None,
    ):
        backend_class = self.BACKENDS[backend]

        if env_directory:
            self.env_directory = str(env_directory)
        elif root_path and env_name:
            self.env_directory = root_path / backend / "envs" / env_name
        else:
            raise Exception(
                "'env_directory' or 'root_path' and 'env_name' should be provided"
            )

        self.backend_instance = backend_class(
            str(self.env_directory), external_executable=str(external_executable)
        )

    def create_environment(self, packages=None, channels=None):
        if channels:
            self.backend_instance.create_environment(packages, channels)
        else:
            self.backend_instance.create_environment(packages)

    def delete_environment(self):
        self.backend_instance.delete_environment()

    def activate(self):
        self.backend_instance.activate_environment()

    def deactivate(self):
        self.backend_instance.deactivate_environment()

    def export_environment(self, export_file_path):
        return self.backend_instance.export_environment(export_file_path)

    def import_environment(self, import_file_path):
        return self.backend_instance.import_environment(import_file_path)

    def install(self, packages=None, channels=None, force=False, capture_output=False):
        if channels:
            return self.backend_instance.install_packages(
                packages=packages,
                channels=channels,
                force=force,
                capture_output=capture_output,
            )
        else:
            return self.backend_instance.install_packages(packages, force=force)

    def uninstall(self, packages, force=False, capture_output=False):
        return self.backend_instance.uninstall_packages(
            packages, force=force, capture_output=capture_output
        )

    def update(self, packages, force=False, capture_output=False):
        return self.backend_instance.update_packages(
            packages, force=force, capture_output=capture_output
        )

    def list(self):
        return self.backend_instance.list_packages()

    @staticmethod
    def list_environments(
        backend=DEFAULT_BACKEND, root_path=DEFAULT_BACKENDS_ROOT_PATH
    ):
        envs_directory = Path(root_path) / backend / "envs"
        environments = {}
        first_environment = True
        print(f"# {backend} environments")
        for env_dir in envs_directory.iterdir():
            if env_dir.is_dir():
                if first_environment:
                    first_environment = False
                environments[env_dir.name] = str(env_dir)
                print(f"{env_dir.name} - {str(env_dir)}")
        else:
            if first_environment:
                print(f"No environments found for {backend} in {root_path}")
        return environments



# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT

import os
import shutil
import subprocess
from pathlib import Path

import requests

from envs_manager.api import EnvManagerInstance

PYPI_API_PACKAGE_INFO_URL = "https://pypi.org/pypi/{package_name}/json"


class VEnvInterface(EnvManagerInstance):
    ID = "venv"

    def _run_command(self, command, capture_output=True):
        run_env = os.environ.copy()
        run_env["PIP_REQUIRE_VIRTUALENV"] = "true"
        if capture_output:
            result = subprocess.run(
                command,
                capture_output=capture_output,
                check=True,
                text=True,
                env=run_env,
            )
        else:
            result = subprocess.run(
                command, stderr=subprocess.PIPE, check=True, text=True, env=run_env
            )
        return result

    def _get_package_info(self, package_name):
        package_info_url = PYPI_API_PACKAGE_INFO_URL.format(package_name=package_name)
        package_info = requests.get(package_info_url).json()
        return package_info

    @property
    def executable_path(self):
        if os.name == "nt":
            executable_path = Path(self.environment_path) / "Scripts" / "python.exe"
        else:
            executable_path = Path(self.environment_path) / "bin" / "python"

        return str(executable_path)

    def validate(self):
        try:
            import venv

            return True
        except ImportError:
            return False

    def create_environment(self, packages=None, channels=None):
        from venv import EnvBuilder

        builder = EnvBuilder(with_pip=True)
        builder.create(self.environment_path)
        if packages:
            try:
                packages.remove("python")
            except ValueError:
                pass
            possible_pythons = [
                package
                for package in packages
                if package.startswith(("python=", "python<", "python>"))
            ]
            for possible_python in possible_pythons:
                packages.remove(possible_python)
            if len(packages) > 0:
                self.install_packages(packages=packages)

    def delete_environment(self):
        shutil.rmtree(self.environment_path)

    def activate_environment(self):
        raise NotImplementedError()

    def deactivate_environment(self):
        raise NotImplementedError()

    def export_environment(self, export_file_path=None):
        try:
            command = [
                self.executable_path,
                "-m",
                "pip",
                "list",
                "--format=freeze",
                "--not-required",
            ]
            result = self._run_command(command)
            if export_file_path:
                with open(export_file_path, "w") as exported_file:
                    exported_file.write(result.stdout)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def import_environment(self, import_file_path):
        self.create_environment()
        try:
            command = [
                self.executable_path,
                "-m",
                "pip",
                "install",
                "-r",
                import_file_path,
            ]
            result = self._run_command(command)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def install_packages(
        self,
        packages,
        channels=None,
        force=False,
        capture_output=False,
    ):
        try:
            command = [self.executable_path, "-m", "pip", "install"] + packages
            result = self._run_command(command, capture_output=capture_output)
            print(result.stdout)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def uninstall_packages(self, packages, force=False, capture_output=False):
        try:
            command = [self.executable_path, "-m", "pip", "uninstall"]
            if force:
                command += ["-y"]
            command += packages
            result = self._run_command(command, capture_output=capture_output)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def update_packages(self, packages, force=False, capture_output=False):
        try:
            command = [self.executable_path, "-m", "pip", "install", "-U"]
            command += packages
            result = self._run_command(command, capture_output=capture_output)
            return (True, result)
        except subprocess.CalledProcessError as error:
            return (False, f"{error.returncode}: {error.stderr}")

    def list_packages(self):
        command = [self.executable_path, "-m", "pip", "list"]
        result = self._run_command(command)
        result_lines = result.stdout.split("\n")

        formatted_packages = {}
        formatted_list = dict(
            environment=self.environment_path, packages=formatted_packages
        )
        for package in result_lines[2:-1]:
            package_name, package_version = package.split()
            package_description = self._get_package_info(package_name)["info"][
                "summary"
            ]
            formatted_package = dict(
                name=package_name,
                version=package_version,
                build=None,
                channel=None,
                description=package_description,
                requested=True,
            )
            formatted_packages[package_name] = formatted_package

        print(result.stdout)
        return formatted_list



# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT

import os
import subprocess

import pytest


SUBCOMMANDS = [
    "",
    "create",
    "delete",
    "activate",
    "deactivate",
    "export",
    "import",
    "install",
    "uninstall",
    "update",
    "list",
    "list-environments",
]

BACKENDS = [
    ("venv", [""], "test_env"),
    (
        "conda-like",
        ["Transaction finished", "Executing transaction: ...working... done"],
        "test_env",
    ),
]


@pytest.mark.parametrize("subcommand", SUBCOMMANDS)
def test_cli_help(subcommand):
    if subcommand:
        subprocess.run(
            ["envs-manager", subcommand, "--help"],
            check=True,
        )
    else:
        subprocess.run(
            ["envs-manager", "--help"],
            check=True,
        )


@pytest.mark.parametrize("backend", BACKENDS)
def test_cli(tmp_path, backend):
    backend_value, create_result, list_env_result = backend
    backends_root_path = tmp_path / "backends"
    backends_root_path.mkdir(parents=True)
    os.environ["BACKENDS_ROOT_PATH"] = str(backends_root_path)

    # Check environment creation
    create_output = subprocess.check_output(
        " ".join(
            ["envs-manager", f"-b={backend_value}", f"-en={list_env_result}", "create"]
        ),
        shell=True,
    )
    assert any(
        [
            expected_create_result in str(create_output)
            for expected_create_result in create_result
        ]
    )

    # Check environment listing
    list_env_output = subprocess.check_output(
        " ".join(
            [
                "envs-manager",
                f"-b={backend_value}",
                f"-en=test_create",
                "list-environments",
            ]
        ),
        shell=True,
    )
    assert list_env_result in str(list_env_output)



# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT

import os
from pathlib import Path
import time
import sys

import pytest

from envs_manager.manager import Manager


TESTS_DIR = Path(__file__).parent.absolute()
ENV_FILES = TESTS_DIR / "env_files"
ENV_FILES_CONDA_LIKE = ENV_FILES / "conda-like-files"
ENV_FILES_VENV = ENV_FILES / "venv-files"
BACKENDS = [
    (
        ("venv", None, None),
        "pip",
        ["packaging==21.0"],
        ["packaging"],
        ["foo"],
        (
            ["Could not find a version that satisfies the requirement foo"],
            ["Could not find a version that satisfies the requirement foo"],
            ["WARNING: Skipping foo as it is not installed"],
        ),
        # Key returned by list call, Number of packages returned, Number of properties returned per package, Package description
        [2, 1, 6, "The PyPA recommended tool for installing Python packages."],
    ),
    (
        ("venv", None, "test_env"),
        "pip",
        ["packaging==21.0"],
        ["packaging"],
        ["foo"],
        (
            ["Could not find a version that satisfies the requirement foo"],
            ["Could not find a version that satisfies the requirement foo"],
            ["WARNING: Skipping foo as it is not installed"],
        ),
        # Key returned by list call, Number of packages returned, Number of properties returned per package, Package description
        [2, 1, 6, "The PyPA recommended tool for installing Python packages."],
    ),
    (
        ("conda-like", os.environ.get("ENV_BACKEND_EXECUTABLE"), None),
        "python",
        ["packaging=21.0"],
        ["packaging"],
        ["foo"],
        (
            ["libmamba Could not solve for environment specs", "PackagesNotFoundError"],
            ["All requested packages already installed", "PackageNotInstalledError"],
            ["Nothing to do", "PackagesNotFoundError"],
        ),
        # Key returned by list call, Number of packages returned, Number of properties returned per package, Package description
        [2, 1, 6, "General purpose programming language"],
    ),
    (
        ("conda-like", os.environ.get("ENV_BACKEND_EXECUTABLE"), "test_env"),
        "python",
        ["packaging=21.0"],
        ["packaging"],
        ["foo"],
        (
            ["libmamba Could not solve for environment specs", "PackagesNotFoundError"],
            ["All requested packages already installed", "PackageNotInstalledError"],
            ["Nothing to do", "PackagesNotFoundError"],
        ),
        # Number of keys returned by list call, Number of packages returned, Number of properties returned per package, Package description
        [2, 1, 6, "General purpose programming language"],
    ),
]


if os.name == "nt":
    BASE_IMPORT_EXPORT_FILENAME = "win"
elif sys.platform.startswith("linux"):
    BASE_IMPORT_EXPORT_FILENAME = "linux"
else:
    BASE_IMPORT_EXPORT_FILENAME = "macos"

IMPORT_EXPORT_BACKENDS = [
    (
        ("venv", None, None),
        str(ENV_FILES_VENV / "venv_import_env.txt"),
        str(ENV_FILES_VENV / "venv_export_env.txt"),
    ),
    (
        ("venv", None, "test_env"),
        str(ENV_FILES_VENV / "venv_import_env.txt"),
        str(ENV_FILES_VENV / "venv_export_env.txt"),
    ),
    (
        ("conda-like", os.environ.get("ENV_BACKEND_EXECUTABLE"), None),
        str(
            ENV_FILES_CONDA_LIKE / f"{BASE_IMPORT_EXPORT_FILENAME}_conda_import_env.yml"
        ),
        str(
            ENV_FILES_CONDA_LIKE / f"{BASE_IMPORT_EXPORT_FILENAME}_conda_export_env.yml"
        ),
    ),
    (
        ("conda-like", os.environ.get("ENV_BACKEND_EXECUTABLE"), "test_env"),
        str(
            ENV_FILES_CONDA_LIKE / f"{BASE_IMPORT_EXPORT_FILENAME}_conda_import_env.yml"
        ),
        str(
            ENV_FILES_CONDA_LIKE / f"{BASE_IMPORT_EXPORT_FILENAME}_conda_export_env.yml"
        ),
    ),
]


def wait_until(condition, interval=0.1, timeout=1, **kwargs):
    start = time.time()
    while not condition(**kwargs) and time.time() - start < timeout:
        time.sleep(interval)
    assert condition(**kwargs)


def package_installed(manager_instance, installed_packages):
    package_list = manager_instance.list()
    for package in installed_packages:
        return package.split("=")[0] in package_list["packages"]


def packages_uninstalled(manager_instance, installed_packages):
    package_list = manager_instance.list()
    for package in installed_packages:
        return package not in package_list["packages"]


def check_packages(manager_instance, package, version):
    packages_list = manager_instance.list()
    return (
        package in packages_list["packages"]
        and packages_list["packages"][package]["version"] == version
    )


@pytest.fixture
def manager_instance(request, tmp_path):
    backend, executable, env_name = request.param
    if not env_name:
        # Passing full env directory
        root_path = None
        envs_directory = tmp_path / "envs"
        env_directory = envs_directory / f"test_{backend}"
        if backend == "conda-like":
            envs_directory.mkdir(parents=True)
        else:
            env_directory.mkdir(parents=True)
    else:
        # Pass root_path and env_name letting manager create env directory path
        root_path = tmp_path
        env_directory = None

    manager_instance = Manager(
        backend,
        root_path=root_path,
        env_name=env_name,
        env_directory=env_directory,
        external_executable=executable,
    )
    yield manager_instance
    manager_instance.delete_environment()


@pytest.mark.parametrize(
    "manager_instance,initial_package,installed_packages,updated_packages,errored_packages,messages,list_info",
    BACKENDS,
    indirect=["manager_instance"],
)
def test_manager_backends(
    manager_instance,
    initial_package,
    installed_packages,
    updated_packages,
    errored_packages,
    messages,
    list_info,
):
    # Create an environment with Python in it
    manager_instance.create_environment(packages=["python==3.10"])

    # List packages and check correct list result dimensions
    initial_list = manager_instance.list()
    assert initial_package in initial_list["packages"]
    (
        list_info_number,
        package_number,
        package_info_number,
        package_description,
    ) = list_info
    assert len(initial_list) == list_info_number
    assert len(initial_list["packages"]) > package_number
    assert len(initial_list["packages"][initial_package]) == package_info_number
    assert (
        initial_list["packages"][initial_package]["description"] == package_description
    )

    # Install a new package in the created environment
    install_result = manager_instance.install(packages=installed_packages, force=True)
    assert install_result[0]

    # Check package was installed
    wait_until(
        package_installed,
        manager_instance=manager_instance,
        installed_packages=installed_packages,
    )

    # Update installed package
    update_result = manager_instance.update(packages=updated_packages, force=True)
    assert update_result[0]

    # Uninstall the new package
    uninstall_result = manager_instance.uninstall(
        packages=installed_packages, force=True
    )
    assert uninstall_result[0]

    # Check package was uninstalled
    wait_until(
        packages_uninstalled,
        manager_instance=manager_instance,
        installed_packages=installed_packages,
    )

    (
        install_error_messages,
        update_warning_messages,
        uninstall_warning_messages,
    ) = messages
    # Try to install unexisting package
    install_error_result, install_error_message = manager_instance.install(
        packages=errored_packages, force=True
    )
    assert not install_error_result
    assert any(
        [
            install_expected_error in install_error_message
            for install_expected_error in install_error_messages
        ]
    )

    # Try to update unexisting package
    update_error_result, update_error_message = manager_instance.update(
        packages=errored_packages, force=True, capture_output=True
    )
    print(update_error_message)
    assert not update_error_result
    assert any(
        [
            update_expected_error in update_error_message
            for update_expected_error in update_warning_messages
        ]
    )

    # Try to uninstall unexisting package
    uninstall_warning_result, uninstall_warning_message = manager_instance.uninstall(
        packages=errored_packages, force=True, capture_output=True
    )
    print(uninstall_warning_message)
    assert uninstall_warning_result
    assert any(
        [
            uninstall_expected_error in uninstall_warning_message.stdout
            or uninstall_expected_error in uninstall_warning_message.stderr
            for uninstall_expected_error in uninstall_warning_messages
        ]
    )


@pytest.mark.parametrize(
    "manager_instance,initial_import_path,expected_export_path",
    IMPORT_EXPORT_BACKENDS,
    indirect=["manager_instance"],
)
def test_manager_backends_import_export(
    manager_instance, initial_import_path, expected_export_path, tmp_path
):
    # Import environment definition
    import_result = manager_instance.import_environment(initial_import_path)
    print(import_result[1])
    assert import_result[0]
    wait_until(
        check_packages,
        manager_instance=manager_instance,
        package="packaging",
        version="21.3",
    )

    # Install packaging 21.0
    channels = None
    if manager_instance.backend_instance.ID == "conda-like":
        channels = ["conda-forge"]
    install_result, message = manager_instance.install(
        ["packaging==21.0"], channels=channels, force=True, capture_output=True
    )
    print(message)
    assert install_result

    # Wait until package is installed
    wait_until(
        check_packages,
        manager_instance=manager_instance,
        package="packaging",
        version="21.0",
    )

    # Export environment and check if is the expected one
    export_path_dir = tmp_path / "exported"
    export_path_dir.mkdir()
    if manager_instance.backend_instance.ID == "venv":
        export_path_file = export_path_dir / "exported_file.txt"
    else:
        export_path_file = export_path_dir / "exported_file.yml"

    export_result = manager_instance.export_environment(str(export_path_file))
    print(export_result[1])
    assert export_result[0]
    assert export_path_file.exists()

    with open(expected_export_path) as expected_export_path_file:
        expected_export = set()
        for expected_line in expected_export_path_file.readlines():
            if (
                "=" in expected_line
                and "ca-certificates" not in expected_line
                and "openssl" not in expected_line
            ):
                expected_export.add(expected_line.strip())

    with open(export_path_file) as generated_export_path_file:
        generated_export = set()
        for generated_line in generated_export_path_file.readlines():
            if (
                "=" in generated_line
                and "ca-certificates" not in generated_line
                and "openssl" not in generated_line
            ):
                generated_export.add(generated_line.strip())

    assert generated_export == expected_export



# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT


class EnvManagerInstance:
    ID = ""

    def __init__(self, environment_path, external_executable=None):
        self.environment_path = environment_path
        self.external_executable = external_executable
        self.executable_variant = None
        assert self.validate(), f"{self.ID} backend unavailable!"

    def validate(self):
        pass

    def create_environment(self, packages=None, channels=None):
        raise NotImplementedError()

    def delete_environment(self):
        raise NotImplementedError()

    def activate_environment(self):
        raise NotImplementedError()

    def deactivate_environment(self):
        raise NotImplementedError()

    def export_environment(self, export_file_path=None):
        raise NotImplementedError()

    def import_environment(self, import_file_path):
        raise NotImplementedError()

    def install_packages(
        self,
        packages,
        channels=None,
        force=False,
        capture_output=False,
    ):
        raise NotImplementedError()

    def uninstall_packages(self, packages, force=False, capture_output=False):
        raise NotImplementedError()

    def update_packages(self, packages, force=False, capture_output=False):
        raise NotImplementedError()

    def list_packages(self):
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022-present Spyder Development Team and envs-manager contributors
#
# SPDX-License-Identifier: MIT

import argparse
import os
from pathlib import Path

from envs_manager.manager import (
    DEFAULT_BACKENDS_ROOT_PATH,
    DEFAULT_BACKEND,
    DEFAULT_ENVS_ROOT_PATH,
    EXTERNAL_EXECUTABLE,
    Manager,
)


def main(args=None):
    for backend in Manager.BACKENDS.keys():
        (DEFAULT_BACKENDS_ROOT_PATH / backend / "envs").mkdir(
            parents=True, exist_ok=True
        )

    parser = argparse.ArgumentParser(
        prog=__name__,
        description="Manage a virtual Python " "environment in a target " "directory.",
    )
    parser.add_argument(
        "-b",
        "--backend",
        default=DEFAULT_BACKEND,
        choices=list(Manager.BACKENDS.keys()),
        help="The implementation to "
        "create/manage a virtual "
        "Python environment in the given "
        "directory.",
    )
    parser.add_argument(
        "-en",
        "--env_name",
        help="The name of a directory where the virtual environment is or will be locate."
        "The environment is/will be located at <BACKENDS_ROOT_PATH>/<ENV_BACKEND>/envs/<env name>.",
    )

    main_subparser = parser.add_subparsers(title="commands", dest="command")

    # Create env
    parser_create = main_subparser.add_parser(
        "create",
        help="Create a virtual Python environment in the target directory.",
    )
    parser_create.add_argument(
        "--packages", nargs="+", help="List of packages to install."
    )
    parser_create.add_argument(
        "--channels", nargs="+", help="List of channels from where to install."
    )

    # Delete env
    parser_delete = main_subparser.add_parser(
        "delete",
        help="Delete a virtual Python environment in the target directory.",
    )

    # Activate env
    parser_activate = main_subparser.add_parser(
        "activate",
        help="Activate the virtual Python environment in the target directory.",
    )

    # Deactivate env
    parser_deactivate = main_subparser.add_parser(
        "deactivate",
        help="Deactivate the virtual Python environment in the target directory.",
    )

    # Export env
    parser_export = main_subparser.add_parser(
        "export",
        help="Export a virtual Python environment in the target directory to a file.",
    )
    parser_export.add_argument(
        "export_file_path", help="File path to export the environment."
    )

    # Import env
    parser_import = main_subparser.add_parser(
        "import",
        help="Import a virtual Python environment in the target directory from a file.",
    )
    parser_import.add_argument(
        "import_file_path",
        help="File path from where to import the environment.",
    )

    # Install packages
    parser_install = main_subparser.add_parser(
        "install",
        help="Install packages in the "
        "virtual Python "
        "environment placed in the "
        "target directory.",
    )
    parser_install.add_argument(
        "packages", nargs="+", help="List of packages to install."
    )
    parser_install.add_argument(
        "--channels", nargs="+", help="List of channels from where to install."
    )

    # Uninstall packages
    parser_uninstall = main_subparser.add_parser(
        "uninstall",
        help="Uninstall packages in the "
        "virtual Python "
        "environment placed in the "
        "target directory.",
    )
    parser_uninstall.add_argument(
        "packages", nargs="+", help="List of packages to uninstall."
    )

    # Update packages
    parser_update = main_subparser.add_parser(
        "update",
        help="Update packages in the "
        "virtual Python "
        "environment placed in the "
        "target directory.",
    )
    parser_update.add_argument(
        "packages", nargs="+", help="List of packages to update."
    )

    # List packages
    parser_list = main_subparser.add_parser(
        "list",
        help="List packages available in the "
        "virtual Python "
        "environment placed in the "
        "target directory.",
    )

    # List environments
    parser_list_environments = main_subparser.add_parser(
        "list-environments",
        help="List discoverable environments available with the current configuration.",
    )

    options = parser.parse_args(args)
    print(options)
    print(f"Using BACKENDS_ROOT_PATH: {DEFAULT_BACKENDS_ROOT_PATH}")
    print(f"Using ENV_BACKEND: {options.backend}")
    print(f"Using ENV_BACKEND_EXECUTABLE: {EXTERNAL_EXECUTABLE}")
    print(f"Default root path to environments: {DEFAULT_ENVS_ROOT_PATH}")
    if options.env_name:
        manager = Manager(
            backend=options.backend,
            env_name=options.env_name,
            root_path=DEFAULT_BACKENDS_ROOT_PATH,
            external_executable=EXTERNAL_EXECUTABLE,
        )
        if options.command == "create":
            manager.create_environment(
                packages=options.packages or ["python"], channels=options.channels
            )
        elif options.command == "delete":
            manager.delete_environment()
        elif options.command == "activate":
            manager.activate()
        elif options.command == "deactivate":
            manager.deactivate()
        elif options.command == "export":
            manager.export_environment(options.export_file_path)
        elif options.command == "import":
            manager.import_environment(options.import_file_path)
        elif options.command == "install":
            manager.install(packages=options.packages)
        elif options.command == "uninstall":
            manager.uninstall(packages=options.packages)
        elif options.command == "update":
            manager.update(packages=options.packages)
        elif options.command == "list":
            manager.list()
    if options.command == "list-environments":
        if options.backend:
            Manager.list_environments(backend=options.backend)
        else:
            Manager.list_environments()
